@@ -202,6 +202,7 @@ struct symt_compiland* symt_new_compiland(struct module* module,
 struct symt_public* symt_new_public(struct module* module, 
                                     struct symt_compiland* compiland,
                                     const char* name,
+                                    BOOL is_function,
                                     unsigned long address, unsigned size)
 {
     struct symt_public* sym;
@@ -217,6 +218,7 @@ struct symt_public* symt_new_public(struct module* module,
         sym->symt.tag      = SymTagPublicSymbol;
         sym->hash_elt.name = pool_strdup(&module->pool, name);
         sym->container     = compiland ? &compiland->symt : NULL;
+        sym->is_function   = is_function;
         sym->address       = address;
         sym->size          = size;
         symt_add_module_ht(module, (struct symt_ht*)sym);
@@ -673,11 +675,16 @@ static void symt_fill_sym_info(struct module_pair* pair,
         }
         break;
     case SymTagPublicSymbol:
-        sym_info->Flags |= SYMFLAG_EXPORT;
-        symt_get_address(sym, &sym_info->Address);
+        {
+            const struct symt_public* pub = (const struct symt_public*)sym;
+            if (pub->is_function)
+                sym_info->Flags |= SYMFLAG_PUBLIC_CODE;
+            else
+                sym_info->Flags |= SYMFLAG_EXPORT;
+            symt_get_address(sym, &sym_info->Address);
+        }
         break;
     case SymTagFunction:
-        sym_info->Flags |= SYMFLAG_FUNCTION;
         symt_get_address(sym, &sym_info->Address);
         break;
     case SymTagThunk:
@@ -843,6 +850,33 @@ static void symt_get_length(struct module* module, const struct symt* symt, ULON
     *size = 0x1000; /* arbitrary value */
 }
 
+/* needed by symt_find_nearest */
+static int symt_get_best_at(struct module* module, int idx_sorttab)
+{
+    ULONG64 ref_addr;
+    int idx_sorttab_orig = idx_sorttab;
+    if (module->addr_sorttab[idx_sorttab]->symt.tag == SymTagPublicSymbol)
+    {
+        symt_get_address(&module->addr_sorttab[idx_sorttab]->symt, &ref_addr);
+        while (idx_sorttab > 0 &&
+               module->addr_sorttab[idx_sorttab]->symt.tag == SymTagPublicSymbol &&
+               !cmp_sorttab_addr(module, idx_sorttab - 1, ref_addr))
+            idx_sorttab--;
+        if (module->addr_sorttab[idx_sorttab]->symt.tag == SymTagPublicSymbol)
+        {
+            idx_sorttab = idx_sorttab_orig;
+            while (idx_sorttab < module->num_sorttab - 1 &&
+                   module->addr_sorttab[idx_sorttab]->symt.tag == SymTagPublicSymbol &&
+                   !cmp_sorttab_addr(module, idx_sorttab + 1, ref_addr))
+                idx_sorttab++;
+        }
+        /* if no better symbol was found restore the original */
+        if (module->addr_sorttab[idx_sorttab]->symt.tag == SymTagPublicSymbol)
+            idx_sorttab = idx_sorttab_orig;
+    }
+    return idx_sorttab;
+}
+
 /* assume addr is in module */
 struct symt_ht* symt_find_nearest(struct module* module, DWORD_PTR addr)
 {
@@ -861,7 +895,12 @@ struct symt_ht* symt_find_nearest(struct module* module, DWORD_PTR addr)
     high = module->num_sorttab;
 
     symt_get_address(&module->addr_sorttab[0]->symt, &ref_addr);
-    if (addr < ref_addr) return NULL;
+    if (addr <= ref_addr)
+    {
+        low = symt_get_best_at(module, 0);
+        return module->addr_sorttab[low];
+    }
+
     if (high)
     {
         symt_get_address(&module->addr_sorttab[high - 1]->symt, &ref_addr);
@@ -884,23 +923,7 @@ struct symt_ht* symt_find_nearest(struct module* module, DWORD_PTR addr)
     /* If found symbol is a public symbol, check if there are any other entries that
      * might also have the same address, but would get better information
      */
-    if (module->addr_sorttab[low]->symt.tag == SymTagPublicSymbol)
-    {
-        symt_get_address(&module->addr_sorttab[low]->symt, &ref_addr);
-        if (low > 0 &&
-            module->addr_sorttab[low - 1]->symt.tag != SymTagPublicSymbol &&
-            !cmp_sorttab_addr(module, low - 1, ref_addr))
-            low--;
-        else if (low < module->num_sorttab - 1 &&
-                 module->addr_sorttab[low + 1]->symt.tag != SymTagPublicSymbol &&
-                 !cmp_sorttab_addr(module, low + 1, ref_addr))
-            low++;
-    }
-    /* finally check that we fit into the found symbol */
-    symt_get_address(&module->addr_sorttab[low]->symt, &ref_addr);
-    if (addr < ref_addr) return NULL;
-    symt_get_length(module, &module->addr_sorttab[low]->symt, &ref_size);
-    if (addr >= ref_addr + ref_size) return NULL;
+    low = symt_get_best_at(module, low);
 
     return module->addr_sorttab[low];
 }
@@ -1229,7 +1252,7 @@ BOOL WINAPI SymFromAddr(HANDLE hProcess, DWORD64 Address,
 
     symt_fill_sym_info(&pair, NULL, &sym->symt, Symbol);
     if (Displacement)
-        *Displacement = Address - Symbol->Address;
+        *Displacement = (Address >= Symbol->Address) ? (Address - Symbol->Address) : (DWORD64)-1;
     return TRUE;
 }
 
@@ -2077,7 +2100,7 @@ BOOL WINAPI SymAddSymbol(HANDLE hProcess, ULONG64 BaseOfDll, PCSTR name,
 {
     WCHAR       nameW[MAX_SYM_NAME];
 
-    MultiByteToWideChar(CP_ACP, 0, name, -1, nameW, sizeof(nameW) / sizeof(WCHAR));
+    MultiByteToWideChar(CP_ACP, 0, name, -1, nameW, ARRAY_SIZE(nameW));
     return SymAddSymbolW(hProcess, BaseOfDll, nameW, addr, size, flags);
 }
 
