@@ -26,11 +26,16 @@
 #include <stdarg.h>
 #include <string.h>
 #include <stdio.h>
+#include "ntstatus.h"
+#define WIN32_NO_STATUS
 #include "windef.h"
 #include "winbase.h"
+#include "wingdi.h"
 #include "ddrawgdi.h"
 #include "wine/winbase16.h"
+#include "winuser.h"
 #include "winternl.h"
+#include "ddk/d3dkmthk.h"
 
 #include "gdi_private.h"
 #include "wine/unicode.h"
@@ -60,6 +65,10 @@ static CRITICAL_SECTION_DEBUG critsect_debug =
 };
 static CRITICAL_SECTION driver_section = { &critsect_debug, -1, 0, 0, 0, 0 };
 
+static HWND (WINAPI *pGetDesktopWindow)(void);
+static INT (WINAPI *pGetSystemMetrics)(INT);
+static DPI_AWARENESS_CONTEXT (WINAPI *pSetThreadDpiAwarenessContext)(DPI_AWARENESS_CONTEXT);
+
 /**********************************************************************
  *	     create_driver
  *
@@ -97,7 +106,7 @@ static const struct gdi_dc_funcs *get_display_driver(void)
     if (!display_driver)
     {
         HMODULE user32 = LoadLibraryA( "user32.dll" );
-        HWND (WINAPI *pGetDesktopWindow)(void) = (void *)GetProcAddress( user32, "GetDesktopWindow" );
+        pGetDesktopWindow = (void *)GetProcAddress( user32, "GetDesktopWindow" );
 
         if (!pGetDesktopWindow() || !display_driver)
         {
@@ -166,6 +175,7 @@ done:
 void CDECL __wine_set_display_driver( HMODULE module )
 {
     struct graphics_driver *driver;
+    HMODULE user32;
 
     if (!(driver = create_driver( module )))
     {
@@ -174,6 +184,10 @@ void CDECL __wine_set_display_driver( HMODULE module )
     }
     if (InterlockedCompareExchangePointer( (void **)&display_driver, driver, NULL ))
         HeapFree( GetProcessHeap(), 0, driver );
+
+    user32 = LoadLibraryA( "user32.dll" );
+    pGetSystemMetrics = (void *)GetProcAddress( user32, "GetSystemMetrics" );
+    pSetThreadDpiAwarenessContext = (void *)GetProcAddress( user32, "SetThreadDpiAwarenessContext" );
 }
 
 
@@ -308,8 +322,8 @@ static INT nulldrv_GetDeviceCaps( PHYSDEV dev, INT cap )
                                          GetDeviceCaps( dev->hdc, LOGPIXELSX ) * 10 );
     case VERTSIZE:        return MulDiv( GetDeviceCaps( dev->hdc, VERTRES ), 254,
                                          GetDeviceCaps( dev->hdc, LOGPIXELSY ) * 10 );
-    case HORZRES:         return 640;
-    case VERTRES:         return 480;
+    case HORZRES:         return pGetSystemMetrics ? pGetSystemMetrics( SM_CXSCREEN ) : 640;
+    case VERTRES:         return pGetSystemMetrics ? pGetSystemMetrics( SM_CYSCREEN ) : 480;
     case BITSPIXEL:       return 32;
     case PLANES:          return 1;
     case NUMBRUSHES:      return -1;
@@ -344,8 +358,28 @@ static INT nulldrv_GetDeviceCaps( PHYSDEV dev, INT cap )
     case SCALINGFACTORX:  return 0;
     case SCALINGFACTORY:  return 0;
     case VREFRESH:        return GetDeviceCaps( dev->hdc, TECHNOLOGY ) == DT_RASDISPLAY ? 1 : 0;
-    case DESKTOPVERTRES:  return GetDeviceCaps( dev->hdc, VERTRES );
-    case DESKTOPHORZRES:  return GetDeviceCaps( dev->hdc, HORZRES );
+    case DESKTOPHORZRES:
+        if (GetDeviceCaps( dev->hdc, TECHNOLOGY ) == DT_RASDISPLAY && pGetSystemMetrics)
+        {
+            DPI_AWARENESS_CONTEXT context;
+            UINT ret;
+            context = pSetThreadDpiAwarenessContext( DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE );
+            ret = pGetSystemMetrics( SM_CXVIRTUALSCREEN );
+            pSetThreadDpiAwarenessContext( context );
+            return ret;
+        }
+        return GetDeviceCaps( dev->hdc, HORZRES );
+    case DESKTOPVERTRES:
+        if (GetDeviceCaps( dev->hdc, TECHNOLOGY ) == DT_RASDISPLAY && pGetSystemMetrics)
+        {
+            DPI_AWARENESS_CONTEXT context;
+            UINT ret;
+            context = pSetThreadDpiAwarenessContext( DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE );
+            ret = pGetSystemMetrics( SM_CYVIRTUALSCREEN );
+            pSetThreadDpiAwarenessContext( context );
+            return ret;
+        }
+        return GetDeviceCaps( dev->hdc, VERTRES );
     case BLTALIGNMENT:    return 0;
     case SHADEBLENDCAPS:  return 0;
     case COLORMGMTCAPS:   return 0;
@@ -1223,3 +1257,12 @@ NTSTATUS WINAPI D3DKMTEscape( const void *pData )
     FIXME("(%p): stub\n", pData);
     return STATUS_NO_MEMORY;
 }
+
+/******************************************************************************
+ *		D3DKMTCloseAdapter [GDI32.@]
+ */
+NTSTATUS WINAPI D3DKMTCloseAdapter( const D3DKMT_CLOSEADAPTER *desc )
+{
+    FIXME("(%p): stub\n", desc);
+    return STATUS_SUCCESS;
+}
