@@ -49,6 +49,10 @@ static int running_under_wine;
 static int winetest_debug;
 static int winetest_report_success;
 
+static POBJECT_TYPE *pExEventObjectType, *pIoFileObjectType, *pPsThreadType;
+
+void WINAPI ObfReferenceObject( void *obj );
+
 extern int CDECL _vsnprintf(char *str, size_t len, const char *format, __ms_va_list argptr);
 
 static void kvprintf(const char *format, __ms_va_list ap)
@@ -122,7 +126,7 @@ static void WINAPIV ok_(const char *file, int line, int condition, const char *m
     __ms_va_end(args);
 }
 
-void vskip_(const char *file, int line, const char *msg, __ms_va_list args)
+static void vskip_(const char *file, int line, const char *msg, __ms_va_list args)
 {
     const char *current_file;
 
@@ -137,7 +141,7 @@ void vskip_(const char *file, int line, const char *msg, __ms_va_list args)
     skipped++;
 }
 
-void WINAPIV win_skip_(const char *file, int line, const char *msg, ...)
+static void WINAPIV win_skip_(const char *file, int line, const char *msg, ...)
 {
     __ms_va_list args;
     __ms_va_start(args, msg);
@@ -174,6 +178,21 @@ static void winetest_end_todo(void)
 #define todo_wine_if(is_todo)   todo_if((is_todo) && running_under_wine)
 #define win_skip(...)           win_skip_(__FILE__, __LINE__, __VA_ARGS__)
 
+static unsigned int strlenW( const WCHAR *str )
+{
+    const WCHAR *s = str;
+    while (*s) s++;
+    return s - str;
+}
+
+static void *kmemcpy(void *dest, const void *src, SIZE_T n)
+{
+    const char *s = src;
+    char *d = dest;
+    while (n--) *d++ = *s++;
+    return dest;
+}
+
 static void *get_proc_address(const char *name)
 {
     UNICODE_STRING name_u;
@@ -289,6 +308,14 @@ static NTSTATUS wait_multiple(ULONG count, void *objs[], WAIT_TYPE wait_type, UL
     return KeWaitForMultipleObjects(count, objs, wait_type, Executive, KernelMode, FALSE, &integer, NULL);
 }
 
+static NTSTATUS wait_single_handle(HANDLE handle, ULONGLONG timeout)
+{
+    LARGE_INTEGER integer;
+
+    integer.QuadPart = timeout;
+    return ZwWaitForSingleObject(handle, FALSE, &integer);
+}
+
 static void run_thread(PKSTART_ROUTINE proc, void *arg)
 {
     OBJECT_ATTRIBUTES attr = {0};
@@ -322,11 +349,13 @@ static void WINAPI mutex_thread(void *arg)
 static void test_sync(void)
 {
     KSEMAPHORE semaphore, semaphore2;
-    KEVENT manual_event, auto_event;
+    KEVENT manual_event, auto_event, *event;
     KTIMER timer;
     LARGE_INTEGER timeout;
+    OBJECT_ATTRIBUTES attr;
     void *objs[2];
     NTSTATUS ret;
+    HANDLE handle;
     int i;
 
     KeInitializeEvent(&manual_event, NotificationEvent, FALSE);
@@ -421,6 +450,30 @@ static void test_sync(void)
     ret = wait_multiple(2, objs, WaitAny, 0);
     ok(ret == 1, "got %#x\n", ret);
 
+    InitializeObjectAttributes(&attr, NULL, OBJ_KERNEL_HANDLE, NULL, NULL);
+    ret = ZwCreateEvent(&handle, SYNCHRONIZE, &attr, NotificationEvent, TRUE);
+    ok(!ret, "ZwCreateEvent failed: %#x\n", ret);
+
+    ret = ObReferenceObjectByHandle(handle, SYNCHRONIZE, *pExEventObjectType, KernelMode, (void **)&event, NULL);
+    ok(!ret, "ObReferenceObjectByHandle failed: %#x\n", ret);
+
+    ret = wait_single(event, 0);
+    ok(ret == 0, "got %#x\n", ret);
+    KeResetEvent(event);
+    ret = wait_single(event, 0);
+    ok(ret == STATUS_TIMEOUT, "got %#x\n", ret);
+    ret = wait_single_handle(handle, 0);
+    ok(ret == STATUS_TIMEOUT, "got %#x\n", ret);
+
+    KeSetEvent(event, 0, FALSE);
+    ret = wait_single(event, 0);
+    ok(ret == 0, "got %#x\n", ret);
+    ret = wait_single_handle(handle, 0);
+    ok(!ret, "got %#x\n", ret);
+
+    ZwClose(handle);
+    ObDereferenceObject(event);
+
     /* test semaphores */
     KeInitializeSemaphore(&semaphore, 0, 5);
 
@@ -636,6 +689,131 @@ static void test_version(void)
     ok(*pNtBuildNumber == build, "Expected build number %u, got %u\n", build, *pNtBuildNumber);
 }
 
+static void WINAPI thread_proc(void *arg)
+{
+    PsTerminateSystemThread(STATUS_SUCCESS);
+}
+
+static void test_ob_reference(const WCHAR *test_path)
+{
+    POBJECT_TYPE (WINAPI *pObGetObjectType)(void*);
+    OBJECT_ATTRIBUTES attr = { sizeof(attr) };
+    HANDLE event_handle, file_handle, file_handle2, thread_handle;
+    FILE_OBJECT *file;
+    void *obj1, *obj2;
+    POBJECT_TYPE obj1_type;
+    UNICODE_STRING pathU;
+    IO_STATUS_BLOCK io;
+    WCHAR *tmp_path;
+    SIZE_T len;
+    NTSTATUS status;
+
+    static const WCHAR tmpW[] = {'.','t','m','p',0};
+
+    pObGetObjectType = get_proc_address("ObGetObjectType");
+    if (!pObGetObjectType)
+        win_skip("ObGetObjectType not found\n");
+
+    InitializeObjectAttributes(&attr, NULL, OBJ_KERNEL_HANDLE, NULL, NULL);
+    status = ZwCreateEvent(&event_handle, SYNCHRONIZE, &attr, NotificationEvent, TRUE);
+    ok(!status, "ZwCreateEvent failed: %#x\n", status);
+
+    len = strlenW(test_path);
+    tmp_path = ExAllocatePool(PagedPool, len * sizeof(WCHAR) + sizeof(tmpW));
+    kmemcpy(tmp_path, test_path, len * sizeof(WCHAR));
+    kmemcpy(tmp_path + len, tmpW, sizeof(tmpW));
+
+    RtlInitUnicodeString(&pathU, tmp_path);
+    attr.ObjectName = &pathU;
+    attr.Attributes = OBJ_KERNEL_HANDLE;
+    status = ZwCreateFile(&file_handle,  DELETE | FILE_WRITE_DATA | SYNCHRONIZE, &attr, &io, NULL, 0, 0, FILE_CREATE,
+                          FILE_DELETE_ON_CLOSE | FILE_SYNCHRONOUS_IO_NONALERT, NULL, 0);
+    ok(!status, "ZwCreateFile failed: %#x\n", status);
+    ExFreePool(tmp_path);
+
+    status = ZwDuplicateObject(NtCurrentProcess(), file_handle, NtCurrentProcess(), &file_handle2,
+                               0, OBJ_KERNEL_HANDLE, DUPLICATE_SAME_ACCESS);
+    ok(!status, "ZwDuplicateObject failed: %#x\n", status);
+
+    InitializeObjectAttributes(&attr, NULL, OBJ_KERNEL_HANDLE, NULL, NULL);
+    status = PsCreateSystemThread(&thread_handle, SYNCHRONIZE, &attr, NULL, NULL, thread_proc, NULL);
+    ok(!status, "PsCreateSystemThread returned: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(NULL, SYNCHRONIZE, *pExEventObjectType, KernelMode, &obj1, NULL);
+    ok(status == STATUS_INVALID_HANDLE, "ObReferenceObjectByHandle failed: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(event_handle, SYNCHRONIZE, *pIoFileObjectType, KernelMode, &obj1, NULL);
+    ok(status == STATUS_OBJECT_TYPE_MISMATCH, "ObReferenceObjectByHandle returned: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(event_handle, SYNCHRONIZE, NULL, KernelMode, &obj1, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+
+    if (pObGetObjectType)
+    {
+        obj1_type = pObGetObjectType(obj1);
+        ok(obj1_type == *pExEventObjectType, "ObGetObjectType returned %p\n", obj1_type);
+    }
+
+    if (sizeof(void *) != 4) /* avoid dealing with fastcall */
+    {
+        ObfReferenceObject(obj1);
+        ObDereferenceObject(obj1);
+    }
+
+    status = ObReferenceObjectByHandle(event_handle, SYNCHRONIZE, *pIoFileObjectType, KernelMode, &obj2, NULL);
+    ok(status == STATUS_OBJECT_TYPE_MISMATCH, "ObReferenceObjectByHandle returned: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(event_handle, SYNCHRONIZE, *pExEventObjectType, KernelMode, &obj2, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+    ok(obj1 == obj2, "obj1 != obj2\n");
+
+    ObDereferenceObject(obj2);
+
+    status = ObReferenceObjectByHandle(event_handle, SYNCHRONIZE, NULL, KernelMode, &obj2, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+    ok(obj1 == obj2, "obj1 != obj2\n");
+
+    ObDereferenceObject(obj2);
+    ObDereferenceObject(obj1);
+
+    status = ObReferenceObjectByHandle(file_handle, SYNCHRONIZE, *pIoFileObjectType, KernelMode, &obj1, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(file_handle2, SYNCHRONIZE, *pIoFileObjectType, KernelMode, &obj2, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+    todo_wine
+    ok(obj1 == obj2, "obj1 != obj2\n");
+
+    file = obj1;
+    ok(file->Type == 5, "Type = %u\n", file->Type);
+
+    ObDereferenceObject(obj1);
+    ObDereferenceObject(obj2);
+
+    status = ObReferenceObjectByHandle(thread_handle, SYNCHRONIZE, *pPsThreadType, KernelMode, &obj1, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+
+    status = ObReferenceObjectByHandle(thread_handle, SYNCHRONIZE, *pPsThreadType, KernelMode, &obj2, NULL);
+    ok(!status, "ObReferenceObjectByHandle failed: %#x\n", status);
+    todo_wine
+    ok(obj1 == obj2, "obj1 != obj2\n");
+
+    ObDereferenceObject(obj1);
+    ObDereferenceObject(obj2);
+
+    status = ZwClose(thread_handle);
+    ok(!status, "ZwClose failed: %#x\n", status);
+
+    status = ZwClose(event_handle);
+    ok(!status, "ZwClose failed: %#x\n", status);
+
+    status = ZwClose(file_handle);
+    ok(!status, "ZwClose failed: %#x\n", status);
+
+    status = ZwClose(file_handle2);
+    ok(!status, "ZwClose failed: %#x\n", status);
+}
+
 static NTSTATUS main_test(DEVICE_OBJECT *device, IRP *irp, IO_STACK_LOCATION *stack, ULONG_PTR *info)
 {
     ULONG length = stack->Parameters.DeviceIoControl.OutputBufferLength;
@@ -647,7 +825,6 @@ static NTSTATUS main_test(DEVICE_OBJECT *device, IRP *irp, IO_STACK_LOCATION *st
 
     if (!buffer)
         return STATUS_ACCESS_VIOLATION;
-
     if (length < sizeof(failures))
         return STATUS_BUFFER_TOO_SMALL;
 
@@ -660,6 +837,15 @@ static NTSTATUS main_test(DEVICE_OBJECT *device, IRP *irp, IO_STACK_LOCATION *st
     attr.Attributes = OBJ_KERNEL_HANDLE; /* needed to be accessible from system threads */
     ZwOpenFile(&okfile, FILE_APPEND_DATA | SYNCHRONIZE, &attr, &io, 0, FILE_SYNCHRONOUS_IO_NONALERT);
 
+    pExEventObjectType = get_proc_address("ExEventObjectType");
+    ok(!!pExEventObjectType, "ExEventObjectType not found\n");
+
+    pIoFileObjectType = get_proc_address("IoFileObjectType");
+    ok(!!pIoFileObjectType, "IofileObjectType not found\n");
+
+    pPsThreadType = get_proc_address("PsThreadType");
+    ok(!!pPsThreadType, "IofileObjectType not found\n");
+
     test_irp_struct(irp, device);
     test_currentprocess();
     test_mdl_map();
@@ -669,9 +855,10 @@ static NTSTATUS main_test(DEVICE_OBJECT *device, IRP *irp, IO_STACK_LOCATION *st
     test_version();
     test_stack_callout();
     test_lookaside_list();
+    test_ob_reference(test_input->path);
 
     /* print process report */
-    if (test_input->winetest_debug)
+    if (winetest_debug)
     {
         kprintf("%04x:ntoskrnl: %d tests executed (%d marked as todo, %d %s), %d skipped.\n",
             PsGetCurrentProcessId(), successes + failures + todo_successes + todo_failures,
