@@ -1084,6 +1084,34 @@ static void test_query_async( IWbemServices *services )
     SysFreeString( query );
 }
 
+static void test_query_semisync( IWbemServices *services )
+{
+    static const WCHAR queryW[] =
+        {'S','E','L','E','C','T',' ','*',' ','F','R','O','M',' ','W','i','n','3','2','_','D','u','m','m','y',0};
+    BSTR wql = SysAllocString( wqlW ), query = SysAllocString( queryW );
+    IEnumWbemClassObject *result;
+    IWbemClassObject *obj;
+    ULONG count;
+    HRESULT hr;
+
+    hr = IWbemServices_ExecQuery( services, wql, query, WBEM_FLAG_RETURN_IMMEDIATELY | WBEM_FLAG_FORWARD_ONLY,
+            NULL, &result );
+    ok( hr == S_OK, "got %08x\n", hr );
+
+    count = 1;
+    obj = (void *)0xdeadbeef;
+    hr = IEnumWbemClassObject_Next( result, -1, 1, &obj, &count );
+todo_wine
+    ok( hr == WBEM_E_INVALID_CLASS, "Unexpected hr %#x.\n", hr );
+    ok( count == 0, "Unexpected count %u.\n", count );
+    ok( obj == (void *)0xdeadbeef, "Got object %p\n", obj );
+
+    IEnumWbemClassObject_Release( result );
+
+    SysFreeString( wql );
+    SysFreeString( query );
+}
+
 static void test_GetNames( IWbemServices *services )
 {
     static const WCHAR queryW[] =
@@ -1206,6 +1234,7 @@ static void test_Win32_OperatingSystem( IWbemServices *services )
     static const WCHAR csdversionW[] = {'C','S','D','V','e','r','s','i','o','n',0};
     static const WCHAR freephysicalmemoryW[] = {'F','r','e','e','P','h','y','s','i','c','a','l','M','e','m','o','r','y',0};
     static const WCHAR nameW[] = {'N','a','m','e',0};
+    static const WCHAR operatingsystemskuW[] = {'O','p','e','r','a','t','i','n','g','S','y','s','t','e','m','S','K','U',0};
     static const WCHAR osproductsuiteW[] = {'O','S','P','r','o','d','u','c','t','S','u','i','t','e',0};
     static const WCHAR ostypeW[] = {'O','S','T','y','p','e',0};
     static const WCHAR suitemaskW[] = {'S','u','i','t','e','M','a','s','k',0};
@@ -1285,6 +1314,18 @@ static void test_Win32_OperatingSystem( IWbemServices *services )
     trace( "name: %s\n", wine_dbgstr_w(V_BSTR( &val )) );
     VariantClear( &val );
 
+    type = 0xdeadbeef;
+    VariantInit( &val );
+    hr = IWbemClassObject_Get( obj, operatingsystemskuW, 0, &val, &type, NULL );
+    ok( hr == S_OK || broken(hr == WBEM_E_NOT_FOUND) /* winxp */, "failed to get operatingsystemsku %08x\n", hr );
+    if (hr == S_OK)
+    {
+        ok( V_VT( &val ) == VT_I4, "unexpected variant type 0x%x\n", V_VT( &val ) );
+        ok( type == CIM_UINT32, "unexpected type 0x%x\n", type );
+        trace( "operatingsystemsku: %08x\n", V_I4( &val ) );
+        VariantClear( &val );
+    }
+
     type = 0xdeadbeef;
     VariantInit( &val );
     hr = IWbemClassObject_Get( obj, osproductsuiteW, 0, &val, &type, NULL );
@@ -1461,6 +1502,7 @@ static void test_Win32_PhysicalMemory( IWbemServices *services )
 {
     static const WCHAR capacityW[] = {'C','a','p','a','c','i','t','y',0};
     static const WCHAR memorytypeW[] = {'M','e','m','o','r','y','T','y','p','e',0};
+    static const WCHAR devicelocatorW[] = {'D','e','v','i','c','e','L','o','c','a','t','o','r',0};
     static const WCHAR queryW[] =
         {'S','E','L','E','C','T',' ','*',' ','F','R','O','M',' ','W','i','n','3','2','_',
          'P','h','y','s','i','c','a','l','M','e','m','o','r','y',0};
@@ -1493,6 +1535,15 @@ static void test_Win32_PhysicalMemory( IWbemServices *services )
         trace( "capacity %s\n", wine_dbgstr_w(V_BSTR( &val )) );
         VariantClear( &val );
 
+        type = 0xdeadbeef;
+        VariantInit( &val );
+        hr = IWbemClassObject_Get( obj, devicelocatorW, 0, &val, &type, NULL );
+        ok( hr == S_OK, "failed to get devicelocator %08x\n", hr );
+        ok( V_VT( &val ) == VT_BSTR, "unexpected variant type 0x%x\n", V_VT( &val ) );
+        ok( type == CIM_STRING, "unexpected type 0x%x\n", type );
+        trace( "devicelocator %s\n", wine_dbgstr_w(V_BSTR( &val )) );
+        VariantClear( &val );
+
         type = 0xdeadbeef;
         VariantInit( &val );
         hr = IWbemClassObject_Get( obj, memorytypeW, 0, &val, &type, NULL );
@@ -1950,6 +2001,7 @@ START_TEST(query)
     test_StdRegProv( services );
     test_notification_query_async( services );
     test_query_async( services );
+    test_query_semisync( services );
     test_GetNames( services );
     test_SystemSecurity( services );
     test_Win32_OperatingSystem( services );
