@@ -44,6 +44,7 @@ enum wined3d_gl_extension
     /* ARB */
     ARB_BASE_INSTANCE,
     ARB_BLEND_FUNC_EXTENDED,
+    ARB_BUFFER_STORAGE,
     ARB_CLEAR_BUFFER_OBJECT,
     ARB_CLEAR_TEXTURE,
     ARB_CLIP_CONTROL,
@@ -55,6 +56,7 @@ enum wined3d_gl_extension
     ARB_CULL_DISTANCE,
     ARB_DEBUG_OUTPUT,
     ARB_DEPTH_BUFFER_FLOAT,
+    ARB_DEPTH_CLAMP,
     ARB_DEPTH_TEXTURE,
     ARB_DERIVATIVE_CONTROL,
     ARB_DRAW_BUFFERS,
@@ -87,14 +89,19 @@ enum wined3d_gl_extension
     ARB_PIXEL_BUFFER_OBJECT,
     ARB_POINT_PARAMETERS,
     ARB_POINT_SPRITE,
+    ARB_POLYGON_OFFSET_CLAMP,
     ARB_PROVOKING_VERTEX,
+    ARB_QUERY_BUFFER_OBJECT,
+    ARB_SAMPLE_SHADING,
     ARB_SAMPLER_OBJECTS,
     ARB_SEAMLESS_CUBE_MAP,
     ARB_SHADER_ATOMIC_COUNTERS,
+    ARB_SHADER_VIEWPORT_LAYER_ARRAY,
     ARB_SHADER_BIT_ENCODING,
     ARB_SHADER_IMAGE_LOAD_STORE,
     ARB_SHADER_IMAGE_SIZE,
     ARB_SHADER_STORAGE_BUFFER_OBJECT,
+    ARB_SHADER_TEXTURE_IMAGE_SAMPLES,
     ARB_SHADER_TEXTURE_LOD,
     ARB_SHADING_LANGUAGE_100,
     ARB_SHADING_LANGUAGE_420PACK,
@@ -150,14 +157,15 @@ enum wined3d_gl_extension
     EXT_BLEND_FUNC_SEPARATE,
     EXT_BLEND_MINMAX,
     EXT_BLEND_SUBTRACT,
-    EXT_DRAW_BUFFERS2,
     EXT_DEPTH_BOUNDS_TEST,
+    EXT_DRAW_BUFFERS2,
     EXT_FOG_COORD,
     EXT_FRAMEBUFFER_BLIT,
     EXT_FRAMEBUFFER_MULTISAMPLE,
     EXT_FRAMEBUFFER_OBJECT,
     EXT_GPU_PROGRAM_PARAMETERS,
     EXT_GPU_SHADER4,
+    EXT_MEMORY_OBJECT,
     EXT_PACKED_DEPTH_STENCIL,
     EXT_PACKED_FLOAT,
     EXT_POINT_PARAMETERS,
