@@ -34,17 +34,10 @@ WINE_DEFAULT_DEBUG_CHANNEL(vulkan);
  */
 #define WINE_VULKAN_ICD_VERSION 4
 
-/* All Vulkan structures use this structure for the first elements. */
-struct wine_vk_structure_header
-{
-    VkStructureType sType;
-    void *pNext;
-};
-
 #define wine_vk_find_struct(s, t) wine_vk_find_struct_((void *)s, VK_STRUCTURE_TYPE_##t)
 static void *wine_vk_find_struct_(void *s, VkStructureType t)
 {
-    struct wine_vk_structure_header *header;
+    VkBaseOutStructure *header;
 
     for (header = s; header; header = header->pNext)
     {
@@ -214,7 +207,21 @@ static struct VkQueue_T *wine_vk_device_alloc_queues(struct VkDevice_T *device,
     return queues;
 }
 
-static VkDeviceGroupDeviceCreateInfo *convert_VkDeviceGroupDeviceCreateInfo(const void *src)
+static void *convert_VkPhysicalDeviceFeatures2(const void *src)
+{
+    const VkPhysicalDeviceFeatures2 *in = src;
+    VkPhysicalDeviceFeatures2 *out;
+
+    if (!(out = heap_alloc(sizeof(*out))))
+        return NULL;
+
+    *out = *in;
+    out->pNext = NULL;
+
+    return out;
+}
+
+static void *convert_VkDeviceGroupDeviceCreateInfo(const void *src)
 {
     const VkDeviceGroupDeviceCreateInfo *in = src;
     VkDeviceGroupDeviceCreateInfo *out;
@@ -225,6 +232,7 @@ static VkDeviceGroupDeviceCreateInfo *convert_VkDeviceGroupDeviceCreateInfo(cons
         return NULL;
 
     *out = *in;
+    out->pNext = NULL;
     if (!(physical_devices = heap_calloc(in->physicalDeviceCount, sizeof(*physical_devices))))
     {
         heap_free(out);
@@ -237,6 +245,40 @@ static VkDeviceGroupDeviceCreateInfo *convert_VkDeviceGroupDeviceCreateInfo(cons
     return out;
 }
 
+static void *convert_VkPhysicalDeviceHostQueryResetFeaturesEXT(const void *src)
+{
+    const VkPhysicalDeviceHostQueryResetFeaturesEXT *in = src;
+    VkPhysicalDeviceHostQueryResetFeaturesEXT *out;
+
+    if (!(out = heap_alloc(sizeof(*out))))
+        return NULL;
+
+    *out = *in;
+    out->pNext = NULL;
+
+    return out;
+}
+
+static void wine_vk_device_free_create_info(VkDeviceCreateInfo *create_info)
+{
+    VkPhysicalDeviceHostQueryResetFeaturesEXT *host_query_reset_features;
+    VkPhysicalDeviceFeatures2 *device_features;
+    VkDeviceGroupDeviceCreateInfo *group_info;
+
+    device_features = wine_vk_find_struct(create_info, PHYSICAL_DEVICE_FEATURES_2);
+    group_info = wine_vk_find_struct(create_info, DEVICE_GROUP_DEVICE_CREATE_INFO);
+    host_query_reset_features = wine_vk_find_struct(create_info, PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES_EXT);
+    create_info->pNext = NULL;
+
+    heap_free(device_features);
+    if (group_info)
+    {
+        heap_free((void *)group_info->pPhysicalDevices);
+        heap_free(group_info);
+    }
+    heap_free(host_query_reset_features);
+}
+
 static VkResult wine_vk_device_convert_create_info(const VkDeviceCreateInfo *src,
         VkDeviceCreateInfo *dst)
 {
@@ -250,9 +292,11 @@ static VkResult wine_vk_device_convert_create_info(const VkDeviceCreateInfo *src
      */
     if (src->pNext)
     {
-        const struct wine_vk_structure_header *header;
+        const VkBaseInStructure *header;
+        VkBaseOutStructure *dst_header;
 
         dst->pNext = NULL;
+        dst_header = (VkBaseOutStructure *)dst;
         for (header = src->pNext; header; header = header->pNext)
         {
             switch (header->sType)
@@ -264,12 +308,25 @@ static VkResult wine_vk_device_convert_create_info(const VkDeviceCreateInfo *src
                     break;
 
                 case VK_STRUCTURE_TYPE_DEVICE_GROUP_DEVICE_CREATE_INFO:
-                    if (!(dst->pNext = convert_VkDeviceGroupDeviceCreateInfo(header)))
-                        return VK_ERROR_OUT_OF_HOST_MEMORY;
+                    if (!(dst_header->pNext = convert_VkDeviceGroupDeviceCreateInfo(header)))
+                        goto err;
+                    dst_header = dst_header->pNext;
+                    break;
+
+                case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2:
+                    if (!(dst_header->pNext = convert_VkPhysicalDeviceFeatures2(header)))
+                        goto err;
+                    dst_header = dst_header->pNext;
+                    break;
+
+                case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES_EXT:
+                    if (!(dst_header->pNext = convert_VkPhysicalDeviceHostQueryResetFeaturesEXT(header)))
+                        goto err;
+                    dst_header = dst_header->pNext;
                     break;
 
                 default:
-                    FIXME("Application requested a linked structure of type %#x.\n", header->sType);
+                    FIXME("Application requested a linked structure of type %u.\n", header->sType);
             }
         }
     }
@@ -291,19 +348,10 @@ static VkResult wine_vk_device_convert_create_info(const VkDeviceCreateInfo *src
     }
 
     return VK_SUCCESS;
-}
 
-static void wine_vk_device_free_create_info(VkDeviceCreateInfo *create_info)
-{
-    VkDeviceGroupDeviceCreateInfo *group_info;
-
-    if ((group_info = wine_vk_find_struct(create_info, DEVICE_GROUP_DEVICE_CREATE_INFO)))
-    {
-        heap_free((void *)group_info->pPhysicalDevices);
-        heap_free(group_info);
-    }
-
-    create_info->pNext = NULL;
+err:
+    wine_vk_device_free_create_info(dst);
+    return VK_ERROR_OUT_OF_HOST_MEMORY;
 }
 
 /* Helper function used for freeing a device structure. This function supports full
@@ -369,7 +417,7 @@ static VkResult wine_vk_instance_convert_create_info(const VkInstanceCreateInfo
      */
     if (src->pNext)
     {
-        const struct wine_vk_structure_header *header;
+        const VkBaseInStructure *header;
 
         for (header = src->pNext; header; header = header->pNext)
         {
@@ -383,7 +431,7 @@ static VkResult wine_vk_instance_convert_create_info(const VkInstanceCreateInfo
                     break;
 
                 default:
-                    FIXME("Application requested a linked structure of type %#x.\n", header->sType);
+                    FIXME("Application requested a linked structure of type %u.\n", header->sType);
             }
         }
     }
@@ -997,13 +1045,13 @@ void WINAPI wine_vkGetDeviceQueue(VkDevice device, uint32_t family_index,
 
 void WINAPI wine_vkGetDeviceQueue2(VkDevice device, const VkDeviceQueueInfo2 *info, VkQueue *queue)
 {
-    const struct wine_vk_structure_header *chain;
     struct VkQueue_T *matching_queue;
+    const VkBaseInStructure *chain;
 
     TRACE("%p, %p, %p\n", device, info, queue);
 
     if ((chain = info->pNext))
-        FIXME("Ignoring a linked structure of type %#x.\n", chain->sType);
+        FIXME("Ignoring a linked structure of type %u.\n", chain->sType);
 
     matching_queue = &device->queues[info->queueFamilyIndex][info->queueIndex];
     if (matching_queue->flags != info->flags)
@@ -1111,7 +1159,7 @@ VkResult WINAPI wine_vkQueueSubmit(VkQueue queue, uint32_t count,
         {
             ERR("Unable to allocate memory for comman buffers!\n");
             res = VK_ERROR_OUT_OF_HOST_MEMORY;
-            goto err;
+            goto done;
         }
 
         for (j = 0; j < num_command_buffers; j++)
@@ -1123,7 +1171,7 @@ VkResult WINAPI wine_vkQueueSubmit(VkQueue queue, uint32_t count,
 
     res = queue->device->funcs.p_vkQueueSubmit(queue->queue, count, submits_host, fence);
 
-err:
+done:
     for (i = 0; i < count; i++)
     {
         heap_free((void *)submits_host[i].pCommandBuffers);
@@ -1228,6 +1276,98 @@ VkResult WINAPI wine_vkEnumeratePhysicalDeviceGroupsKHR(VkInstance instance,
             instance->funcs.p_vkEnumeratePhysicalDeviceGroupsKHR, count, properties);
 }
 
+void WINAPI wine_vkGetPhysicalDeviceExternalFenceProperties(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalFenceInfo *fence_info, VkExternalFenceProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, fence_info, properties);
+    properties->exportFromImportedHandleTypes = 0;
+    properties->compatibleHandleTypes = 0;
+    properties->externalFenceFeatures = 0;
+}
+
+void WINAPI wine_vkGetPhysicalDeviceExternalFencePropertiesKHR(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalFenceInfo *fence_info, VkExternalFenceProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, fence_info, properties);
+    properties->exportFromImportedHandleTypes = 0;
+    properties->compatibleHandleTypes = 0;
+    properties->externalFenceFeatures = 0;
+}
+
+void WINAPI wine_vkGetPhysicalDeviceExternalBufferProperties(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalBufferInfo *buffer_info, VkExternalBufferProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, buffer_info, properties);
+    memset(&properties->externalMemoryProperties, 0, sizeof(properties->externalMemoryProperties));
+}
+
+void WINAPI wine_vkGetPhysicalDeviceExternalBufferPropertiesKHR(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalBufferInfo *buffer_info, VkExternalBufferProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, buffer_info, properties);
+    memset(&properties->externalMemoryProperties, 0, sizeof(properties->externalMemoryProperties));
+}
+
+VkResult WINAPI wine_vkGetPhysicalDeviceImageFormatProperties2(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceImageFormatInfo2 *format_info, VkImageFormatProperties2 *properties)
+{
+    VkExternalImageFormatProperties *external_image_properties;
+    VkResult res;
+
+    TRACE("%p, %p, %p\n", phys_dev, format_info, properties);
+
+    res = thunk_vkGetPhysicalDeviceImageFormatProperties2(phys_dev, format_info, properties);
+
+    if ((external_image_properties = wine_vk_find_struct(properties, EXTERNAL_IMAGE_FORMAT_PROPERTIES)))
+    {
+        VkExternalMemoryProperties *p = &external_image_properties->externalMemoryProperties;
+        p->externalMemoryFeatures = 0;
+        p->exportFromImportedHandleTypes = 0;
+        p->compatibleHandleTypes = 0;
+    }
+
+    return res;
+}
+
+VkResult WINAPI wine_vkGetPhysicalDeviceImageFormatProperties2KHR(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceImageFormatInfo2 *format_info, VkImageFormatProperties2 *properties)
+{
+    VkExternalImageFormatProperties *external_image_properties;
+    VkResult res;
+
+    TRACE("%p, %p, %p\n", phys_dev, format_info, properties);
+
+    res = thunk_vkGetPhysicalDeviceImageFormatProperties2KHR(phys_dev, format_info, properties);
+
+    if ((external_image_properties = wine_vk_find_struct(properties, EXTERNAL_IMAGE_FORMAT_PROPERTIES)))
+    {
+        VkExternalMemoryProperties *p = &external_image_properties->externalMemoryProperties;
+        p->externalMemoryFeatures = 0;
+        p->exportFromImportedHandleTypes = 0;
+        p->compatibleHandleTypes = 0;
+    }
+
+    return res;
+}
+
+void WINAPI wine_vkGetPhysicalDeviceExternalSemaphoreProperties(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalSemaphoreInfo *semaphore_info, VkExternalSemaphoreProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, semaphore_info, properties);
+    properties->exportFromImportedHandleTypes = 0;
+    properties->compatibleHandleTypes = 0;
+    properties->externalSemaphoreFeatures = 0;
+}
+
+void WINAPI wine_vkGetPhysicalDeviceExternalSemaphorePropertiesKHR(VkPhysicalDevice phys_dev,
+        const VkPhysicalDeviceExternalSemaphoreInfo *semaphore_info, VkExternalSemaphoreProperties *properties)
+{
+    TRACE("%p, %p, %p\n", phys_dev, semaphore_info, properties);
+    properties->exportFromImportedHandleTypes = 0;
+    properties->compatibleHandleTypes = 0;
+    properties->externalSemaphoreFeatures = 0;
+}
+
 BOOL WINAPI DllMain(HINSTANCE hinst, DWORD reason, void *reserved)
 {
     TRACE("%p, %u, %p\n", hinst, reason, reserved);
