@@ -70,39 +70,39 @@ struct pipe_end
 
 struct pipe_server
 {
-    struct pipe_end      pipe_end;   /* common header for pipe_client and pipe_server */
-    struct list          entry;      /* entry in named pipe servers list */
+    struct pipe_end      pipe_end;   /* common header for both pipe ends */
+    struct list          entry;      /* entry in named pipe listeners list */
     unsigned int         options;    /* pipe options */
     struct async_queue   listen_q;   /* listen queue */
 };
 
-struct pipe_client
-{
-    struct pipe_end      pipe_end;   /* common header for pipe_client and pipe_server */
-    unsigned int         flags;      /* file flags */
-};
-
 struct named_pipe
 {
     struct object       obj;         /* object header */
-    unsigned int        flags;
+    int                 message_mode;
     unsigned int        sharing;
     unsigned int        maxinstances;
     unsigned int        outsize;
     unsigned int        insize;
     unsigned int        instances;
     timeout_t           timeout;
-    struct list         servers;     /* list of servers using this pipe */
+    struct list         listeners;   /* list of servers listening on this pipe */
     struct async_queue  waiters;     /* list of clients waiting to connect */
 };
 
 struct named_pipe_device
 {
     struct object       obj;         /* object header */
-    struct fd          *fd;          /* pseudo-fd for ioctls */
     struct namespace   *pipes;       /* named pipe namespace */
 };
 
+struct named_pipe_device_file
+{
+    struct object             obj;         /* object header */
+    struct fd                *fd;          /* pseudo-fd for ioctls */
+    struct named_pipe_device *device;      /* named pipe device */
+};
+
 static void named_pipe_dump( struct object *obj, int verbose );
 static unsigned int named_pipe_map_access( struct object *obj, unsigned int access );
 static int named_pipe_link_name( struct object *obj, struct object_name *name, struct object *parent );
@@ -128,13 +128,13 @@ static const struct object_ops named_pipe_ops =
     named_pipe_link_name,         /* link_name */
     default_unlink_name,          /* unlink_name */
     named_pipe_open_file,         /* open_file */
+    no_kernel_obj_list,           /* get_kernel_obj_list */
     no_close_handle,              /* close_handle */
     named_pipe_destroy            /* destroy */
 };
 
 /* common server and client pipe end functions */
 static void pipe_end_destroy( struct object *obj );
-static struct object_type *pipe_end_get_type( struct object *obj );
 static enum server_fd_type pipe_end_get_fd_type( struct fd *fd );
 static struct fd *pipe_end_get_fd( struct object *obj );
 static struct security_descriptor *pipe_end_get_sd( struct object *obj );
@@ -156,7 +156,7 @@ static const struct object_ops pipe_server_ops =
 {
     sizeof(struct pipe_server),   /* size */
     pipe_server_dump,             /* dump */
-    pipe_end_get_type,            /* get_type */
+    file_get_type,                /* get_type */
     add_queue,                    /* add_queue */
     remove_queue,                 /* remove_queue */
     default_fd_signaled,          /* signaled */
@@ -170,6 +170,7 @@ static const struct object_ops pipe_server_ops =
     no_link_name,                 /* link_name */
     NULL,                         /* unlink_name */
     no_open_file,                 /* open_file */
+    no_kernel_obj_list,           /* get_kernel_obj_list */
     fd_close_handle,              /* close_handle */
     pipe_server_destroy           /* destroy */
 };
@@ -195,9 +196,9 @@ static int pipe_client_ioctl( struct fd *fd, ioctl_code_t code, struct async *as
 
 static const struct object_ops pipe_client_ops =
 {
-    sizeof(struct pipe_client),   /* size */
+    sizeof(struct pipe_end),      /* size */
     pipe_client_dump,             /* dump */
-    pipe_end_get_type,            /* get_type */
+    file_get_type,                /* get_type */
     add_queue,                    /* add_queue */
     remove_queue,                 /* remove_queue */
     default_fd_signaled,          /* signaled */
@@ -211,6 +212,7 @@ static const struct object_ops pipe_client_ops =
     no_link_name,                 /* link_name */
     NULL,                         /* unlink_name */
     no_open_file,                 /* open_file */
+    no_kernel_obj_list,           /* get_kernel_obj_list */
     fd_close_handle,              /* close_handle */
     pipe_end_destroy              /* destroy */
 };
@@ -232,14 +234,11 @@ static const struct fd_ops pipe_client_fd_ops =
 
 static void named_pipe_device_dump( struct object *obj, int verbose );
 static struct object_type *named_pipe_device_get_type( struct object *obj );
-static struct fd *named_pipe_device_get_fd( struct object *obj );
 static struct object *named_pipe_device_lookup_name( struct object *obj,
     struct unicode_str *name, unsigned int attr );
 static struct object *named_pipe_device_open_file( struct object *obj, unsigned int access,
                                                    unsigned int sharing, unsigned int options );
 static void named_pipe_device_destroy( struct object *obj );
-static enum server_fd_type named_pipe_device_get_fd_type( struct fd *fd );
-static int named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, struct async *async );
 
 static const struct object_ops named_pipe_device_ops =
 {
@@ -251,7 +250,7 @@ static const struct object_ops named_pipe_device_ops =
     NULL,                             /* signaled */
     no_satisfied,                     /* satisfied */
     no_signal,                        /* signal */
-    named_pipe_device_get_fd,         /* get_fd */
+    no_get_fd,                        /* get_fd */
     no_map_access,                    /* map_access */
     default_get_sd,                   /* get_sd */
     default_set_sd,                   /* set_sd */
@@ -259,23 +258,53 @@ static const struct object_ops named_pipe_device_ops =
     directory_link_name,              /* link_name */
     default_unlink_name,              /* unlink_name */
     named_pipe_device_open_file,      /* open_file */
-    fd_close_handle,                  /* close_handle */
+    no_kernel_obj_list,               /* get_kernel_obj_list */
+    no_close_handle,                  /* close_handle */
     named_pipe_device_destroy         /* destroy */
 };
 
+static void named_pipe_device_file_dump( struct object *obj, int verbose );
+static struct fd *named_pipe_device_file_get_fd( struct object *obj );
+static int named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, struct async *async );
+static enum server_fd_type named_pipe_device_file_get_fd_type( struct fd *fd );
+static void named_pipe_device_file_destroy( struct object *obj );
+
+static const struct object_ops named_pipe_device_file_ops =
+{
+    sizeof(struct named_pipe_device_file),   /* size */
+    named_pipe_device_file_dump,             /* dump */
+    no_get_type,                             /* get_type */
+    add_queue,                               /* add_queue */
+    remove_queue,                            /* remove_queue */
+    default_fd_signaled,                     /* signaled */
+    no_satisfied,                            /* satisfied */
+    no_signal,                               /* signal */
+    named_pipe_device_file_get_fd,           /* get_fd */
+    default_fd_map_access,                   /* map_access */
+    default_get_sd,                          /* get_sd */
+    default_set_sd,                          /* set_sd */
+    no_lookup_name,                          /* lookup_name */
+    no_link_name,                            /* link_name */
+    NULL,                                    /* unlink_name */
+    no_open_file,                            /* open_file */
+    no_kernel_obj_list,                      /* get_kernel_obj_list */
+    fd_close_handle,                         /* close_handle */
+    named_pipe_device_file_destroy           /* destroy */
+};
+
 static const struct fd_ops named_pipe_device_fd_ops =
 {
-    default_fd_get_poll_events,       /* get_poll_events */
-    default_poll_event,               /* poll_event */
-    named_pipe_device_get_fd_type,    /* get_fd_type */
-    no_fd_read,                       /* read */
-    no_fd_write,                      /* write */
-    no_fd_flush,                      /* flush */
-    default_fd_get_file_info,         /* get_file_info */
-    no_fd_get_volume_info,            /* get_volume_info */
-    named_pipe_device_ioctl,          /* ioctl */
-    default_fd_queue_async,           /* queue_async */
-    default_fd_reselect_async         /* reselect_async */
+    default_fd_get_poll_events,              /* get_poll_events */
+    default_poll_event,                      /* poll_event */
+    named_pipe_device_file_get_fd_type,      /* get_fd_type */
+    no_fd_read,                              /* read */
+    no_fd_write,                             /* write */
+    no_fd_flush,                             /* flush */
+    default_fd_get_file_info,                /* get_file_info */
+    no_fd_get_volume_info,                   /* get_volume_info */
+    named_pipe_device_ioctl,                 /* ioctl */
+    default_fd_queue_async,                  /* queue_async */
+    default_fd_reselect_async                /* reselect_async */
 };
 
 static void named_pipe_dump( struct object *obj, int verbose )
@@ -302,27 +331,20 @@ static void pipe_server_dump( struct object *obj, int verbose )
 
 static void pipe_client_dump( struct object *obj, int verbose )
 {
-    struct pipe_client *client = (struct pipe_client *) obj;
+    struct pipe_end *client = (struct pipe_end *) obj;
     assert( obj->ops == &pipe_client_ops );
-    fprintf( stderr, "Named pipe client server=%p\n", client->pipe_end.connection );
+    fprintf( stderr, "Named pipe client server=%p\n", client->connection );
 }
 
 static void named_pipe_destroy( struct object *obj)
 {
     struct named_pipe *pipe = (struct named_pipe *) obj;
 
-    assert( list_empty( &pipe->servers ) );
+    assert( list_empty( &pipe->listeners ) );
     assert( !pipe->instances );
     free_async_queue( &pipe->waiters );
 }
 
-static struct object_type *pipe_end_get_type( struct object *obj )
-{
-    static const WCHAR name[] = {'F','i','l','e'};
-    static const struct unicode_str str = { name, sizeof(name) };
-    return get_object_type( &str );
-}
-
 static struct fd *pipe_end_get_fd( struct object *obj )
 {
     struct pipe_end *pipe_end = (struct pipe_end *) obj;
@@ -419,7 +441,8 @@ static void pipe_server_destroy( struct object *obj )
 
     assert( pipe->instances );
     if (!--pipe->instances) unlink_named_object( &pipe->obj );
-    list_remove( &server->entry );
+    if (server->pipe_end.state == FILE_PIPE_LISTENING_STATE)
+        list_remove( &server->entry );
 
     free_async_queue( &server->listen_q );
     pipe_end_destroy( obj );
@@ -437,12 +460,6 @@ static struct object_type *named_pipe_device_get_type( struct object *obj )
     return get_object_type( &str );
 }
 
-static struct fd *named_pipe_device_get_fd( struct object *obj )
-{
-    struct named_pipe_device *device = (struct named_pipe_device *)obj;
-    return (struct fd *)grab_object( device->fd );
-}
-
 static struct object *named_pipe_device_lookup_name( struct object *obj, struct unicode_str *name,
                                                      unsigned int attr )
 {
@@ -463,22 +480,26 @@ static struct object *named_pipe_device_lookup_name( struct object *obj, struct
 static struct object *named_pipe_device_open_file( struct object *obj, unsigned int access,
                                                    unsigned int sharing, unsigned int options )
 {
-    return grab_object( obj );
+    struct named_pipe_device_file *file;
+
+    if (!(file = alloc_object( &named_pipe_device_file_ops ))) return NULL;
+    file->device = (struct named_pipe_device *)grab_object( obj );
+    if (!(file->fd = alloc_pseudo_fd( &named_pipe_device_fd_ops, obj, options )))
+    {
+        release_object( file );
+        return NULL;
+    }
+    allow_fd_caching( file->fd );
+    return &file->obj;
 }
 
 static void named_pipe_device_destroy( struct object *obj )
 {
     struct named_pipe_device *device = (struct named_pipe_device*)obj;
     assert( obj->ops == &named_pipe_device_ops );
-    if (device->fd) release_object( device->fd );
     free( device->pipes );
 }
 
-static enum server_fd_type named_pipe_device_get_fd_type( struct fd *fd )
-{
-    return FD_TYPE_DEVICE;
-}
-
 struct object *create_named_pipe_device( struct object *root, const struct unicode_str *name )
 {
     struct named_pipe_device *dev;
@@ -487,8 +508,7 @@ struct object *create_named_pipe_device( struct object *root, const struct unico
         get_error() != STATUS_OBJECT_NAME_EXISTS)
     {
         dev->pipes = NULL;
-        if (!(dev->fd = alloc_pseudo_fd( &named_pipe_device_fd_ops, &dev->obj, 0 )) ||
-            !(dev->pipes = create_namespace( 7 )))
+        if (!(dev->pipes = create_namespace( 7 )))
         {
             release_object( dev );
             dev = NULL;
@@ -497,6 +517,32 @@ struct object *create_named_pipe_device( struct object *root, const struct unico
     return &dev->obj;
 }
 
+static void named_pipe_device_file_dump( struct object *obj, int verbose )
+{
+    struct named_pipe_device_file *file = (struct named_pipe_device_file *)obj;
+
+    fprintf( stderr, "File on named pipe device %p\n", file->device );
+}
+
+static struct fd *named_pipe_device_file_get_fd( struct object *obj )
+{
+    struct named_pipe_device_file *file = (struct named_pipe_device_file *)obj;
+    return (struct fd *)grab_object( file->fd );
+}
+
+static enum server_fd_type named_pipe_device_file_get_fd_type( struct fd *fd )
+{
+    return FD_TYPE_DEVICE;
+}
+
+static void named_pipe_device_file_destroy( struct object *obj )
+{
+    struct named_pipe_device_file *file = (struct named_pipe_device_file*)obj;
+    assert( obj->ops == &named_pipe_device_file_ops );
+    if (file->fd) release_object( file->fd );
+    release_object( file->device );
+}
+
 static int pipe_end_flush( struct fd *fd, struct async *async )
 {
     struct pipe_end *pipe_end = get_fd_user( fd );
@@ -607,7 +653,7 @@ static void pipe_end_get_file_info( struct fd *fd, obj_handle_t handle, unsigned
             }
 
             if (!(pipe_info = set_reply_data_size( sizeof(*pipe_info) ))) return;
-            pipe_info->NamedPipeType = (pipe_end->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE) != 0;
+            pipe_info->NamedPipeType = pipe->message_mode;
             switch (pipe->sharing)
             {
             case FILE_SHARE_READ:
@@ -845,7 +891,7 @@ static int pipe_end_write( struct fd *fd, struct async *async, file_pos_t pos )
         return 0;
     }
 
-    if (!(pipe_end->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE) && !get_req_data_size()) return 1;
+    if (!pipe_end->pipe->message_mode && !get_req_data_size()) return 1;
 
     iosb = async_get_iosb( async );
     message = queue_message( pipe_end->connection, iosb );
@@ -908,7 +954,7 @@ static int pipe_end_peek( struct pipe_end *pipe_end )
         avail += message->iosb->in_size - message->read_pos;
     reply_size = min( reply_size, avail );
 
-    if (avail && (pipe_end->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE))
+    if (avail && pipe_end->pipe->message_mode)
     {
         message = LIST_ENTRY( list_head(&pipe_end->message_queue), struct pipe_message, entry );
         message_length = message->iosb->in_size - message->read_pos;
@@ -948,8 +994,7 @@ static int pipe_end_transceive( struct pipe_end *pipe_end, struct async *async )
         return 0;
     }
 
-    if ((pipe_end->flags & (NAMED_PIPE_MESSAGE_STREAM_WRITE | NAMED_PIPE_MESSAGE_STREAM_READ))
-        != (NAMED_PIPE_MESSAGE_STREAM_WRITE | NAMED_PIPE_MESSAGE_STREAM_READ))
+    if (!(pipe_end->flags & NAMED_PIPE_MESSAGE_STREAM_READ))
     {
         set_error( STATUS_INVALID_READ_MODE );
         return 0;
@@ -1037,7 +1082,10 @@ static int pipe_server_ioctl( struct fd *fd, ioctl_code_t code, struct async *as
         switch(server->pipe_end.state)
         {
         case FILE_PIPE_LISTENING_STATE:
+            break;
         case FILE_PIPE_DISCONNECTED_STATE:
+            server->pipe_end.state = FILE_PIPE_LISTENING_STATE;
+            list_add_tail( &server->pipe_end.pipe->listeners, &server->entry );
             break;
         case FILE_PIPE_CONNECTED_STATE:
             set_error( STATUS_PIPE_CONNECTED );
@@ -1047,7 +1095,6 @@ static int pipe_server_ioctl( struct fd *fd, ioctl_code_t code, struct async *as
             return 0;
         }
 
-        server->pipe_end.state = FILE_PIPE_LISTENING_STATE;
         queue_async( &server->listen_q, async );
         async_wake_up( &server->pipe_end.pipe->waiters, STATUS_SUCCESS );
         set_error( STATUS_PENDING );
@@ -1082,7 +1129,7 @@ static int pipe_server_ioctl( struct fd *fd, ioctl_code_t code, struct async *as
 
 static int pipe_client_ioctl( struct fd *fd, ioctl_code_t code, struct async *async )
 {
-    struct pipe_client *client = get_fd_user( fd );
+    struct pipe_end *client = get_fd_user( fd );
 
     switch(code)
     {
@@ -1091,7 +1138,7 @@ static int pipe_client_ioctl( struct fd *fd, ioctl_code_t code, struct async *as
         return 0;
 
     default:
-        return pipe_end_ioctl( &client->pipe_end, code, async );
+        return pipe_end_ioctl( client, code, async );
     }
 }
 
@@ -1121,8 +1168,9 @@ static struct pipe_server *create_pipe_server( struct named_pipe *pipe, unsigned
     init_pipe_end( &server->pipe_end, pipe, pipe_flags, pipe->insize );
     server->pipe_end.state = FILE_PIPE_LISTENING_STATE;
     server->pipe_end.server_pid = get_process_id( current->process );
+    init_async_queue( &server->listen_q );
 
-    list_add_head( &pipe->servers, &server->entry );
+    list_add_tail( &pipe->listeners, &server->entry );
     if (!(server->pipe_end.fd = alloc_pseudo_fd( &pipe_server_fd_ops, &server->pipe_end.obj, options )))
     {
         release_object( server );
@@ -1130,57 +1178,34 @@ static struct pipe_server *create_pipe_server( struct named_pipe *pipe, unsigned
     }
     allow_fd_caching( server->pipe_end.fd );
     set_fd_signaled( server->pipe_end.fd, 1 );
-    init_async_queue( &server->listen_q );
+    async_wake_up( &pipe->waiters, STATUS_SUCCESS );
     return server;
 }
 
-static struct pipe_client *create_pipe_client( unsigned int flags, struct named_pipe *pipe,
-                                               data_size_t buffer_size, unsigned int options )
+static struct pipe_end *create_pipe_client( struct named_pipe *pipe, data_size_t buffer_size, unsigned int options )
 {
-    struct pipe_client *client;
+    struct pipe_end *client;
 
     client = alloc_object( &pipe_client_ops );
     if (!client)
         return NULL;
 
-    client->flags = flags;
-    init_pipe_end( &client->pipe_end, pipe, pipe->flags, buffer_size );
-    client->pipe_end.state = FILE_PIPE_CONNECTED_STATE;
-    client->pipe_end.client_pid = get_process_id( current->process );
+    init_pipe_end( client, pipe, 0, buffer_size );
+    client->state = FILE_PIPE_CONNECTED_STATE;
+    client->client_pid = get_process_id( current->process );
 
-    client->pipe_end.fd = alloc_pseudo_fd( &pipe_client_fd_ops, &client->pipe_end.obj, options );
-    if (!client->pipe_end.fd)
+    client->fd = alloc_pseudo_fd( &pipe_client_fd_ops, &client->obj, options );
+    if (!client->fd)
     {
         release_object( client );
         return NULL;
     }
-    allow_fd_caching( client->pipe_end.fd );
-    set_fd_signaled( client->pipe_end.fd, 1 );
+    allow_fd_caching( client->fd );
+    set_fd_signaled( client->fd, 1 );
 
     return client;
 }
 
-static struct pipe_server *find_available_server( struct named_pipe *pipe )
-{
-    struct pipe_server *server;
-
-    /* look for pipe servers that are listening */
-    LIST_FOR_EACH_ENTRY( server, &pipe->servers, struct pipe_server, entry )
-    {
-        if (server->pipe_end.state == FILE_PIPE_LISTENING_STATE && async_queued( &server->listen_q ))
-            return (struct pipe_server *)grab_object( server );
-    }
-
-    /* fall back to pipe servers that are idle */
-    LIST_FOR_EACH_ENTRY( server, &pipe->servers, struct pipe_server, entry )
-    {
-        if (server->pipe_end.state == FILE_PIPE_LISTENING_STATE )
-            return (struct pipe_server *)grab_object( server );
-    }
-
-    return NULL;
-}
-
 static int named_pipe_link_name( struct object *obj, struct object_name *name, struct object *parent )
 {
     struct named_pipe_device *dev = (struct named_pipe_device *)parent;
@@ -1200,35 +1225,35 @@ static struct object *named_pipe_open_file( struct object *obj, unsigned int acc
 {
     struct named_pipe *pipe = (struct named_pipe *)obj;
     struct pipe_server *server;
-    struct pipe_client *client;
+    struct pipe_end *client;
     unsigned int pipe_sharing;
 
-    if (!(server = find_available_server( pipe )))
+    if (list_empty( &pipe->listeners ))
     {
         set_error( STATUS_PIPE_NOT_AVAILABLE );
         return NULL;
     }
+    server = LIST_ENTRY( list_head( &pipe->listeners ), struct pipe_server, entry );
 
     pipe_sharing = pipe->sharing;
     if (((access & GENERIC_READ) && !(pipe_sharing & FILE_SHARE_READ)) ||
         ((access & GENERIC_WRITE) && !(pipe_sharing & FILE_SHARE_WRITE)))
     {
         set_error( STATUS_ACCESS_DENIED );
-        release_object( server );
         return NULL;
     }
 
-    if ((client = create_pipe_client( options, pipe, pipe->outsize, options )))
+    if ((client = create_pipe_client( pipe, pipe->outsize, options )))
     {
         async_wake_up( &server->listen_q, STATUS_SUCCESS );
         server->pipe_end.state = FILE_PIPE_CONNECTED_STATE;
-        server->pipe_end.connection = &client->pipe_end;
-        client->pipe_end.connection = &server->pipe_end;
-        server->pipe_end.client_pid = client->pipe_end.client_pid;
-        client->pipe_end.server_pid = server->pipe_end.server_pid;
+        server->pipe_end.connection = client;
+        client->connection = &server->pipe_end;
+        server->pipe_end.client_pid = client->client_pid;
+        client->server_pid = server->pipe_end.server_pid;
+        list_remove( &server->entry );
     }
-    release_object( server );
-    return &client->pipe_end.obj;
+    return &client->obj;
 }
 
 static int named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, struct async *async )
@@ -1242,7 +1267,6 @@ static int named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, struct asy
             const FILE_PIPE_WAIT_FOR_BUFFER *buffer = get_req_data();
             data_size_t size = get_req_data_size();
             struct named_pipe *pipe;
-            struct pipe_server *server;
             struct unicode_str name;
             timeout_t when;
 
@@ -1256,19 +1280,16 @@ static int named_pipe_device_ioctl( struct fd *fd, ioctl_code_t code, struct asy
             name.len = (buffer->NameLength / sizeof(WCHAR)) * sizeof(WCHAR);
             if (!(pipe = open_named_object( &device->obj, &named_pipe_ops, &name, 0 ))) return 0;
 
-            if (!(server = find_available_server( pipe )))
+            if (list_empty( &pipe->listeners ))
             {
                 queue_async( &pipe->waiters, async );
                 when = buffer->TimeoutSpecified ? buffer->Timeout.QuadPart : pipe->timeout;
                 async_set_timeout( async, when, STATUS_IO_TIMEOUT );
-                release_object( pipe );
                 set_error( STATUS_PENDING );
-                return 1;
             }
 
-            release_object( server );
             release_object( pipe );
-            return 0;
+            return 1;
         }
 
     default:
@@ -1316,12 +1337,12 @@ DECL_HANDLER(create_named_pipe)
         /* initialize it if it didn't already exist */
         pipe->instances = 0;
         init_async_queue( &pipe->waiters );
-        list_init( &pipe->servers );
+        list_init( &pipe->listeners );
         pipe->insize = req->insize;
         pipe->outsize = req->outsize;
         pipe->maxinstances = req->maxinstances;
         pipe->timeout = req->timeout;
-        pipe->flags = req->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE;
+        pipe->message_mode = (req->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE) != 0;
         pipe->sharing = req->sharing;
         if (sd) default_set_sd( &pipe->obj, sd, OWNER_SECURITY_INFORMATION |
                                                 GROUP_SECURITY_INFORMATION |
@@ -1373,14 +1394,18 @@ DECL_HANDLER(set_named_pipe_info)
         if (!pipe_end) return;
     }
 
-    if ((req->flags & ~(NAMED_PIPE_MESSAGE_STREAM_READ | NAMED_PIPE_NONBLOCKING_MODE)) ||
-            ((req->flags & NAMED_PIPE_MESSAGE_STREAM_READ) && !(pipe_end->pipe->flags & NAMED_PIPE_MESSAGE_STREAM_WRITE)))
+    if (!pipe_end->pipe)
+    {
+        set_error( STATUS_PIPE_DISCONNECTED );
+    }
+    else if ((req->flags & ~(NAMED_PIPE_MESSAGE_STREAM_READ | NAMED_PIPE_NONBLOCKING_MODE)) ||
+            ((req->flags & NAMED_PIPE_MESSAGE_STREAM_READ) && !pipe_end->pipe->message_mode))
     {
         set_error( STATUS_INVALID_PARAMETER );
     }
     else
     {
-        pipe_end->flags = pipe_end->pipe->flags | req->flags;
+        pipe_end->flags = req->flags;
     }
 
     release_object( pipe_end );
